"""Lextab module."""

# lextab.py. This file automatically created by PLY (version 3.11). Don't edit!
_tabversion = "3.10"
_lextokens = set(
    (
        "AND",
        "AT",
        "BACKSLASH",
        "BREAK",
        "CASE",
        "CATCH",
        "CCONST",
        "CLASS",
        "COLON",
        "COMMA",
        "COMMENTCLOSE",
        "CONDITIONAL1",
        "CONST",
        "CONTINUE",
        "DEFAULT",
        "DELETE",
        "DIVEQUAL",
        "DIVIDE",
        "DO",
        "DOCSTRINGOPEN",
        "DOLLAR",
        "DQOUTE",
        "ELSE",
        "EQ",
        "EQQ",
        "EQUALS",
        "EXTENDS",
        "FCONST",
        "FOR",
        "FUNCTION",
        "GE",
        "GT",
        "ICONST",
        "ID",
        "IF",
        "IN",
        "LAND",
        "LBRACE",
        "LBRACKET",
        "LE",
        "LNOT",
        "LOR",
        "LPAREN",
        "LSHIFT",
        "LT",
        "MINUS",
        "MINUSEQUAL",
        "MINUSMINUS",
        "MOD",
        "MODEQUAL",
        "NE",
        "NEQ",
        "NEW",
        "OR",
        "PERIOD",
        "PLUS",
        "PLUSEQUAL",
        "PLUSPLUS",
        "RBRACE",
        "RBRACKET",
        "RETURN",
        "RPAREN",
        "RSHIFT",
        "SCONST",
        "SEMI",
        "SQOUTE",
        "STATIC",
        "SWITCH",
        "THROW",
        "TIMES",
        "TIMESEQUAL",
        "TRY",
        "TYPEOF",
        "VAR",
        "WHILE",
        "WITH",
        "XOR",
    )
)
_lexreflags = 64
_lexliterals = ""
_lexstateinfo = {"INITIAL": "inclusive"}
_lexstatere = {
    "INITIAL": [
        (
            "(?P<t_NEWLINE>\\n+)|(?P<t_ID>[A-Za-z_]+[\\w_]*)|(?P<t_comment>(/\\*( |\\*\\*)(.|\\n)*?\\*/)|(//.*))|(?P<t_DOCSTRINGOPEN>/\\*\\*[ ]+)|(?P<t_preprocessor>\\#(.)*?\\n)|(?P<t_FCONST>((\\d+)(\\.\\d+)(e(\\+|-)?(\\d+))? | (\\d+)e(\\+|-)?(\\d+))([lL]|[fF])?)|(?P<t_ICONST>\\d+([uU]|[lL]|[uU][lL]|[lL][uU])?)|(?P<t_CCONST>\\'([^\\'\\\\\\n]|(\\\\.)|\\\\\\n)*?\\')|(?P<t_SCONST>\\\"([^\\\"\\\\\\n]|(\\\\.)|\\\\\\n)*?\\\")|(?P<t_LOR>\\|\\|)|(?P<t_PLUSPLUS>\\+\\+)|(?P<t_COMMENTCLOSE>\\*/)|(?P<t_EQQ>===)|(?P<t_NEQ>!==)|(?P<t_PLUSEQUAL>\\+=)|(?P<t_TIMESEQUAL>\\*=)|(?P<t_BACKSLASH>\\\\)|(?P<t_CONDITIONAL1>\\?)|(?P<t_DIVEQUAL>/=)|(?P<t_DOLLAR>\\$)|(?P<t_EQ>==)|(?P<t_GE>>=)|(?P<t_LAND>&&)|(?P<t_LBRACE>\\{)|(?P<t_LBRACKET>\\[)|(?P<t_LE><=)|(?P<t_LPAREN>\\()|(?P<t_LSHIFT><<)|(?P<t_MINUSEQUAL>-=)|(?P<t_MINUSMINUS>--)|(?P<t_MODEQUAL>%=)|(?P<t_NE>!=)|(?P<t_OR>\\|)|(?P<t_PERIOD>\\.)|(?P<t_PLUS>\\+)|(?P<t_RBRACE>\\})|(?P<t_RBRACKET>\\])|(?P<t_RPAREN>\\))|(?P<t_RSHIFT>>>)|(?P<t_TIMES>\\*)|(?P<t_XOR>\\^)|(?P<t_AND>&)|(?P<t_AT>@)|(?P<t_COLON>:)|(?P<t_COMMA>,)|(?P<t_DIVIDE>/)|(?P<t_DQOUTE>\")|(?P<t_EQUALS>=)|(?P<t_GT>>)|(?P<t_LNOT>!)|(?P<t_LT><)|(?P<t_MINUS>-)|(?P<t_MOD>%)|(?P<t_SEMI>;)|(?P<t_SQOUTE>')",
            [
                None,
                ("t_NEWLINE", "NEWLINE"),
                ("t_ID", "ID"),
                ("t_comment", "comment"),
                None,
                None,
                None,
                None,
                ("t_DOCSTRINGOPEN", "DOCSTRINGOPEN"),
                ("t_preprocessor", "preprocessor"),
                None,
                (None, "FCONST"),
                None,
                None,
                None,
                None,
                None,
                None,
                None,
                None,
                None,
                None,
                (None, "ICONST"),
                None,
                (None, "CCONST"),
                None,
                None,
                (None, "SCONST"),
                None,
                None,
                (None, "LOR"),
                (None, "PLUSPLUS"),
                (None, "COMMENTCLOSE"),
                (None, "EQQ"),
                (None, "NEQ"),
                (None, "PLUSEQUAL"),
                (None, "TIMESEQUAL"),
                (None, "BACKSLASH"),
                (None, "CONDITIONAL1"),
                (None, "DIVEQUAL"),
                (None, "DOLLAR"),
                (None, "EQ"),
                (None, "GE"),
                (None, "LAND"),
                (None, "LBRACE"),
                (None, "LBRACKET"),
                (None, "LE"),
                (None, "LPAREN"),
                (None, "LSHIFT"),
                (None, "MINUSEQUAL"),
                (None, "MINUSMINUS"),
                (None, "MODEQUAL"),
                (None, "NE"),
                (None, "OR"),
                (None, "PERIOD"),
                (None, "PLUS"),
                (None, "RBRACE"),
                (None, "RBRACKET"),
                (None, "RPAREN"),
                (None, "RSHIFT"),
                (None, "TIMES"),
                (None, "XOR"),
                (None, "AND"),
                (None, "AT"),
                (None, "COLON"),
                (None, "COMMA"),
                (None, "DIVIDE"),
                (None, "DQOUTE"),
                (None, "EQUALS"),
                (None, "GT"),
                (None, "LNOT"),
                (None, "LT"),
                (None, "MINUS"),
                (None, "MOD"),
                (None, "SEMI"),
                (None, "SQOUTE"),
            ],
        )
    ]
}
_lexstateignore = {"INITIAL": " \r\t\x0c"}
_lexstateerrorf = {"INITIAL": "t_error"}
_lexstateeoff = {}



"""Test_flsqlite module."""
import unittest
from pineboolib.loader.main import init_testing, finish_testing
from pineboolib.plugins.sql import flmysql_myisam


class TestFLSqlite(unittest.TestCase):
    """TestFLSqlite Class."""

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        init_testing()

    def test_basic_1(self) -> None:
        """Basics test 1."""

        driver = flmysql_myisam.FLMYSQL_MYISAM()

        self.assertEqual(driver.formatValueLike("bool", "true", False), "=0")
        self.assertEqual(driver.formatValueLike("date", "27-01-2020", True), " LIKE '%%2020-01-27'")

        self.assertEqual(driver.formatValue("bool", "false", True), "False")
        self.assertEqual(driver.formatValue("time", "", True), "")

        self.assertEqual(driver.setType("String", 20), "CHAR(20)")
        self.assertEqual(driver.setType("sTring", 0), "CHAR")
        self.assertEqual(driver.setType("Double"), "DECIMAL")
        self.assertEqual(driver.setType("Bool"), "BOOL")
        self.assertEqual(driver.setType("DATE"), "DATE")
        self.assertEqual(driver.setType("pixmap"), "MEDIUMTEXT")
        self.assertEqual(driver.setType("bytearray"), "LONGBLOB")
        self.assertEqual(driver.setType("timestamp"), "TIMESTAMP")

    def test_basic_2(self) -> None:
        """Basics test 2."""
        from pineboolib.application.database import pnsqlcursor

        cursor = pnsqlcursor.PNSqlCursor("fltest")
        sql = (
            "CREATE TABLE fltest (id INT UNSIGNED PRIMARY KEY,string_field CHAR(255) NULL,date_field DATE NULL,time_field TIME NULL,"
            + "double_field DECIMAL(13,7) NULL,bool_field BOOL NULL,uint_field INT UNSIGNED NULL,bloqueo BOOL NOT NULL"
            + ",empty_relation CHAR(15) NULL,int_field INTEGER NULL)"
            + " ENGINE=MyISAM DEFAULT CHARACTER SET = utf8 COLLATE = utf8_bin"
        )

        driver = flmysql_myisam.FLMYSQL_MYISAM()
        self.assertEqual(sql, driver.sqlCreateTable(cursor.metadata(), False))

    def test_basic_3(self) -> None:
        """Basic test 3."""

        driver = flmysql_myisam.FLMYSQL_MYISAM()
        self.assertEqual(driver.decodeSqlType("char"), "string")
        self.assertEqual(driver.decodeSqlType("int unsigned"), "uint")
        self.assertEqual(driver.decodeSqlType("int"), "int")
        self.assertEqual(driver.decodeSqlType("date"), "date")
        self.assertEqual(driver.decodeSqlType("mediumtext"), "stringlist")
        self.assertEqual(driver.decodeSqlType("tinyint"), "bool")
        self.assertEqual(driver.decodeSqlType("decimal"), "double")
        self.assertEqual(driver.decodeSqlType("longblob"), "bytearray")
        self.assertEqual(driver.decodeSqlType("time"), "time")
        self.assertEqual(driver.decodeSqlType("timestamp"), "timestamp")

    @classmethod
    def tearDownClass(cls) -> None:
        """Ensure test clear all data."""
        finish_testing()



"""
Module for garbage collector checks.
"""

from typing import Any, Callable, List

from pineboolib.core.utils import logging

from pineboolib import core
import weakref
import threading
import time

import gc

LOGGER = logging.get_logger(__name__)


def check_delete(obj_: Any, name: str, force: bool = False) -> bool:
    """Delete a object."""

    if not core.DISABLE_CHECK_MEMORY_LEAKS or force:
        check_gc_referrers(obj_.__class__.__name__, weakref.ref(obj_), name)
    return True


def check_gc_referrers(typename: Any, w_obj: Callable, name: str) -> None:
    """
    Check if any variable is getting out of control.

    Great for checking and tracing memory leaks.
    """

    def checkfn() -> None:
        # time.sleep(2)
        list_: List[str] = []
        try:
            if w_obj is not None:
                gc.collect()
                obj = w_obj()
                if not obj:
                    return
                # TODO: Si ves el mensaje a continuación significa que "algo" ha dejado
                # ..... alguna referencia a un formulario (o similar) que impide que se destruya
                # ..... cuando se deja de usar. Causando que los connects no se destruyan tampoco
                # ..... y que se llamen referenciando al código antiguo y fallando.
                for ref in gc.get_referrers(obj):
                    if "<frame" in str(repr(ref)):
                        continue

                    elif isinstance(ref, dict):
                        for key, value in ref.items():
                            if (
                                key in core.PROXY_ACTIONS_DICT.keys()
                            ):  # Falso positivo. Esto es el listado de los hilos.
                                continue
                            if value is obj:
                                list_.append(
                                    "(%s.%s -> %s (%s)" % (ref.__class__.__name__, key, name, ref)
                                )
                        # print(" - dict:", repr(x), gc.get_referrers(ref))
                    else:
                        list_.append(
                            "(%s) %s.%s -> %s (%s)"
                            % (ref.__class__.__name__, type(ref), str(repr(ref)), name, ref)
                        )
                        # print(" - obj:", repr(ref), [x for x in dir(ref) if getattr(ref, x) is obj])
                if list_:
                    LOGGER.warning(
                        "HINT: %d Objetos referenciando %r::%r (%r) :",
                        len(list_),
                        typename,
                        obj,
                        name,
                    )
                    for item in list_:
                        LOGGER.warning(item)

        except Exception as error:  # noqa : F841
            LOGGER.warning("Error cleaning %r::%r (%r) :", typename, obj, name)

    threading.Thread(target=checkfn).start()


def check_active_threads(full: bool = False) -> None:
    """Check active threads."""

    current_thread_ids = [thread.ident for thread in threading.enumerate()]
    proxy_keys = core.PROXY_ACTIONS_DICT.keys()
    if full:
        LOGGER.warning("CHECK ACTIVE THREADS")
        # LOGGER.warning("Active threads: %s" % (current_thread_ids))
    for thread_id in list(proxy_keys):
        if thread_id not in current_thread_ids:
            if full:
                LOGGER.warning("Deleting thread %s" % (thread_id))
            script_names_list = core.PROXY_ACTIONS_DICT[thread_id]
            for script_name in list(script_names_list):
                if full:
                    LOGGER.warning("Deleting action %s from thread %s" % (script_name, thread_id))
                delete_proxy_thread(thread_id, script_name)

            core.PROXY_ACTIONS_DICT[thread_id].clear()
            del core.PROXY_ACTIONS_DICT[thread_id]


def delete_proxy_thread(id_thread: int, script_name: str) -> None:
    """Delete actions from thread."""

    from pineboolib.application import qsadictmodules

    action_obj = qsadictmodules.QSADictModules.from_project(script_name)
    if hasattr(action_obj, "loaded_obj"):
        if id_thread in action_obj.loaded_obj.keys():
            obj_ = action_obj.loaded_obj[id_thread]
            action_obj.loaded_obj[id_thread] = None
            del action_obj.loaded_obj[id_thread]
            # quitar script_name de la lista core.PROXY_ACTIONS_DICT[thread_id]

            if hasattr(obj_, "iface"):
                LOGGER.info("Deleting iface from %s, %s" % (script_name, obj_.iface))

                obj_.iface = None

            check_delete(obj_, "proxy.%s" % script_name)


def register_script_name(script_name: str) -> None:
    """Register script name."""
    if not core.DISABLE_CHECK_MEMORY_LEAKS:
        id_thread: int = threading.current_thread().ident  # type: ignore [assignment]
        if id_thread not in core.PROXY_ACTIONS_DICT.keys():
            core.PROXY_ACTIONS_DICT[id_thread] = []

        script_name = script_name if script_name != "sys" else "sys_module"

        if script_name not in core.PROXY_ACTIONS_DICT[id_thread]:
            core.PROXY_ACTIONS_DICT[id_thread].append(script_name)  # type: ignore [union-attr]


def periodic_gc(interval: int = 60) -> None:
    """Periodic cleaning task."""
    while True:
        LOGGER.warning("Starting periodic GC every %s seconds" % interval)
        check_active_threads(True)
        LOGGER.warning("Next periodic GC in %s seconds" % interval)
        time.sleep(interval)
